--------------------------------------------------
-- Mir AI v1.2.2
--------------------------------------------------
MIRAI_VER = 122

require(Config.MODULE_CONST)
require(Config.MODULE_UTIL)
require(Config.MODULE_DB_PASS)
require(Config.MODULE_PATROL)
require(Config.MODULE_SELECTED_MOD) -- 3rd party changes
require(Config.MODULE_MERCENARY_MOD)
require(Config.MODULE_MERCENARY_SK)
--require(Config.MODULE_ELLYE_ADDON) -- Add On by Ellye

--------------------------------------------------
-- State
--------------------------------------------------
-- from default AI
IDLE_ST              = 0
FOLLOW_ST            = 1
CHASE_ST             = 2
ATTACK_ST            = 3
MOVE_CMD_ST          = 4
STOP_CMD_ST          = 5
ATTACK_OBJECT_CMD_ST = 6
ATTACK_AREA_CMD_ST   = 7
PATROL_CMD_ST        = 8
HOLD_CMD_ST          = 9
SKILL_OBJECT_CMD_ST  = 10
SKILL_AREA_CMD_ST    = 11
FOLLOW_CMD_ST        = 12

-- Mir AI addings
EVADE_ST             = 50
BUGPOSI_ST           = 51
EVADE_COMEBACK       = 2

-- Mir AI "Output movements" (they notify the user about some events):
OUTMOVE_NEWFRIEND_ST = 60 -- The choosen player has been added to the friend-list
OUTMOVE_DELFRIEND_ST = 61 -- " has been removed

--------------------------------------------------
-- Global variable (dafault AI)
--------------------------------------------------
MyEnemy      = 0
MyDestX      = 0
MyDestY      = 0
MyPatrolX    = 0
MyPatrolY    = 0
ResCmdList   = List.new()
MyID         = 0
MySkill      = 0
MySkillLevel = 0

--------------------------------------------------
-- Global variable (extra)
--------------------------------------------------

-- Init (please look at AI() function, init section)
Initialized = false
Private_Srv = false
OwnerID     = 0
LRASID      = 0
StartupTime = 0
isCircling  = false

-- Current status
CircleDir       = AAI_CIRC_MAXSTEP -- Current circle-move direction
CircleBlockTime = 0 -- Time when position started to not change (obstacle encountered?)
CircleBlockStep = 0 -- 0 = all ok; 1 = obstacle? Check a second; 2 = obstacle detected
Evade_ComeBackTime = 0
vMercX          = 0 --\_ I'm really moving?
vMercY          = 0 --/
vMercHP        = 0 ---- hit while chasing?
vOwnerX         = 0 --\_ used in follow at once
vOwnerY         = 0 --/
OwnerX          = 0
OwnerY          = 0
MyX             = 0
MyY             = 0
IdleStartTime   = 0
AtkStartTime    = 0
LastLRAtkTime   = 0 -- time of last long ranged attack (see OnCHASE_ST)
LastLRAtkID     = 0
AtkSkillDoneCount = 0
OwnerMoveTime   = 0 -- see follow st

HTact = {}
HTact.Behav     = -1
HTact.Skill     = -1
HTact.Level     = 0
HTact.Alche     = 0

-- Blocked movements
AltAngle        = 0
BlockTime       = 0
BlockFound      = 0
BlockCount      = 0
BlockedPathToEnemyID = 0
OwnerBlocked    = 0
LstOwnerX       = 0
LstOwnerY       = 0

BugPosiTime     = 0

-- Output movements
OutMoveTime     = 0
OutMoveTarget   = 0

CanDoKS = 0

-----------------------------------------------------
-- Variaveis globais de status
-----------------------------------------------------
Self_HP     = 0 -- HP atual do homunculus/assistente
Self_Max_HP = 0 -- HP mximo do homunculus/assistente
Self_Per_HP = 0 -- % de HP do homunculus/assistente
Self_SP     = 0 -- SP atual do homunculus/assistente
Self_Max_SP = 0 -- SP mximo do homunculus/assistente
Self_Per_SP = 0 -- % de SP do homunculus/assistente
Self_Motion = 0 -- Postura do homunculus/assistente 
Mast_HP     = 0 -- HP atual do mestre
Mast_Max_HP = 0 -- HP mximo do mestre
Mast_Per_HP = 0 -- % de HP do mestre
Mast_Motion = 0 -- Postura do mestre
Scre_Actors = 0 -- Lista de atores visiveis na tela
-----------------------------------------------------

--------------------------------------------------
-- Skills
--------------------------------------------------
CastDelayEnd = 0

--------------------------------------------------
AddDelaySkill = 1000

DELAY_SLOW_POWER = 3000 -- for slow_power skill mode

--------------------------------------------------
-- ############ COMMAND PROCESS ###########
--------------------------------------------------

--------------------------------------------------
function OnMOVE_CMD(x, y)
--------------------------------------------------
	Log("OnMOVE_CMD")
	isCircling = false
	if (x == MyDestX and y == MyDestY and MOTION_MOVE == GetV(V_MOTION, MyID)) then
		return
	end

	-- are you adding/removing a friend? ----------
	local actors = GetActors()
	for i,v in ipairs(actors) do
		if (v ~= OwnerID and v ~= MyID) then
			if (0 == IsMonster(v)) then
				local PlayerX, PlayerY = GetV(V_POSITION, v)
				if (PlayerX == x) and (math.abs(PlayerY-y) <=1) then
					FriendList_Switch(v) -- update friend list
					OutMoveTime = GetTick() -- notify user
					OutMoveTarget = v
					if isNotFriend(v) then
						MyState = OUTMOVE_DELFRIEND_ST
					else
						MyState = OUTMOVE_NEWFRIEND_ST
					end
					MyDestX = x
					MyDestY = y
					MyEnemy = 0
					MySkill = 0
					return
				end
			end
		end
	end

	if (math.abs(x-MyX) + math.abs(y-MyY) > 15) then
		List.pushleft (ResCmdList,{MOVE_CMD,x,y})
		x = math.floor((x+MyX)/2)
		y = math.floor((y+MyY)/2)
	end

	Move(MyID, x, y)

	MyState = MOVE_CMD_ST
	MyDestX = x
	MyDestY = y
	MyEnemy = 0
	MySkill = 0
end

--------------------------------------------------
function OnSTOP_CMD()
--------------------------------------------------
	Log("OnSTOP_CMD")
	isCircling = false
	if (GetV(V_MOTION,MyID) ~= MOTION_STAND) then
		StopHere("Stop request by user")
	end
end

--------------------------------------------------
function OnATTACK_OBJECT_CMD(id)
--------------------------------------------------
	Log("OnATTACK_OBJECT_CMD")
	isCircling = false
	MySkill = 0
	MyEnemy = id
	CanDoKS = MyEnemy
	vOwnerX, vOwnerY = OwnerX, OwnerY
	MyState = CHASE_ST
end

--------------------------------------------------
function OnATTACK_AREA_CMD(x, y)
--------------------------------------------------
	Log("OnATTACK_AREA_CMD")
	isCircling = false
	if (x ~= MyDestX or y ~= MyDestY or MOTION_MOVE ~= GetV(V_MOTION, MyID)) then
		Move(MyID, x, y)
	end
	MyDestX = x
	MyDestY = y
	MyEnemy = 0
	MyState = ATTACK_AREA_CMD_ST
end

--------------------------------------------------
function OnPATROL_CMD(x, y)
--------------------------------------------------
	Log ("OnPATROL_CMD")
	isCircling = false
	MyPatrolX , MyPatrolY = MyX, MyY
	MyDestX = x
	MyDestY = y
	Move(MyID,x,y)
	MyState = PATROL_CMD_ST
end

--------------------------------------------------
function OnHOLD_CMD()
--------------------------------------------------
	Log("OnHOLD_CMD")
	isCircling = false
	MyDestX = 0
	MyDestY = 0
	MyEnemy = 0
	MyState = HOLD_CMD_ST
end

--------------------------------------------------
function OnSKILL_OBJECT_CMD(level, skill, id)
--------------------------------------------------
	Log("OnSKILL_OBJECT_CMD")
	if  (skill == 8221)                      or
	   ((skill >= 8226) and (skill <= 8232)) or
	   ((skill >= 8235) and (skill <= 8236)) then
		SkillObject(MyID, level, skill, id)
		return
	end
	isCircling = false
	MySkillLevel = level
	MySkill = skill
	MyEnemy = id
	CanDoKS = MyEnemy
	vOwnerX, vOwnerY = OwnerX, OwnerY
	MyState = CHASE_ST
end

--------------------------------------------------
function OnSKILL_AREA_CMD(level, skill, x, y)
--------------------------------------------------
	Log("OnSKILL_AREA_CMD")
	isCircling = false
	Move(MyID,x,y)
	MyDestX = x
	MyDestY = y
	MySkillLevel = level
	MySkill = skill
	MyState = SKILL_AREA_CMD_ST
end

--------------------------------------------------
function OnFOLLOW_CMD()
--------------------------------------------------
	isCircling = false
	if (MyState ~= FOLLOW_CMD_ST) then
		local file = io.open(F_FOLLOW_LCK, "wb")
		if file then file:close() end
		MoveToOwner(MyID)
		MyState = FOLLOW_CMD_ST
		MyDestX, MyDestY = OwnerX, OwnerY
		MyEnemy = 0
		MySkill = 0
		Log("OnFOLLOW_CMD")
	else
		if FileExists(F_FOLLOW_LCK) then
			os.remove(F_FOLLOW_LCK)
		end
		MyState = IDLE_ST
		IdleStartTime = GetTick()
		MyEnemy = 0
		MySkill = 0
		Log("FOLLOW_CMD_ST --> IDLE_ST")
	end
end

--------------------------------------------------
function ProcessCommand(msg)
--------------------------------------------------
	if 	(msg[1] == MOVE_CMD) then
		OnMOVE_CMD(msg[2], msg[3])
		Log("MOVE_CMD")
	elseif (msg[1] == STOP_CMD) then
		OnSTOP_CMD()
		Log("STOP_CMD")
	elseif (msg[1] == ATTACK_OBJECT_CMD) then
		OnATTACK_OBJECT_CMD(msg[2])
		Log("ATTACK_OBJECT_CMD")
	elseif (msg[1] == ATTACK_AREA_CMD) then
		OnATTACK_AREA_CMD(msg[2], msg[3])
		Log("ATTACK_AREA_CMD")
	elseif (msg[1] == PATROL_CMD) then
		OnPATROL_CMD(msg[2], msg[3])
		Log("PATROL_CMD")
	elseif (msg[1] == HOLD_CMD) then
		OnHOLD_CMD()
		Log("HOLD_CMD")
	elseif (msg[1] == SKILL_OBJECT_CMD) then
		OnSKILL_OBJECT_CMD(msg[2], msg[3], msg[4], msg[5])
		Log("SKILL_OBJECT_CMD")
	elseif (msg[1] == SKILL_AREA_CMD) then
		OnSKILL_AREA_CMD(msg[2], msg[3], msg[4], msg[5])
		Log ("SKILL_AREA_CMD")
	elseif (msg[1] == FOLLOW_CMD) then
		OnFOLLOW_CMD()
		Log("FOLLOW_CMD")
	end
end

--------------------------------------------------
-- ############# STATE PROCESS ############
--------------------------------------------------

--------------------------------------------------
function OnIDLE_ST()
--------------------------------------------------
	Log("OnIDLE_ST")

	local cmd = List.popleft(ResCmdList)
	if (cmd ~= nil) then
		ProcessCommand(cmd)
		return
	end

	AtkStartTime = 0
	AtkSkillDoneCount = 0
	HTact.Behav = -1
	HTact.Skill = -1

	local CurrTime = GetTick()

	OwnerMotion = GetV(V_MOTION, OwnerID)

	if (Config[Self].LONG_RANGE_SHOOTER ~= true) and (Config[Self].CIRCLE_ON_IDLE > 0) and (isCircling == false) then
		if (Self_HP == Self_Max_HP) and (Self_SP == Self_Max_SP) then
			if (OwnerMotion == MOTION_STAND or OwnerMotion == MOTION_SIT) and isCloseToOwner() then
				isCircling = true
			end
		end
	end

	-- Is there a new target? -> chase_st
	if CurrTime - StartupTime > 600 then
		local NextTarget = GetMyNextTarget(Self_Per_HP)		
		if (NextTarget ~=0) and (not ((NextTarget == MyID) or (NextTarget == OwnerID) or (Friends[NextTarget] ~= nil))) then
			isCircling = false
			MyEnemy = NextTarget
			local Ex, Ey = GetV(V_POSITION, MyEnemy)
			if (math.abs(Ex - MyX) <= 1) and (math.abs(Ey - MyY) <= 1) then
				MyState = ATTACK_ST
				Attack(MyID, MyEnemy)
				AtkStartTime = 0
				Log(string.format("[IDLE_ST -> ATTACK_ST] Target(%d) already near, switch directly to combat", MyEnemy))
			else
				MyState = CHASE_ST
				Log(string.format("[IDLE_ST -> CHASE_ST] Intercepting new target(%d)", MyEnemy))
			end
			vOwnerX, vOwnerY = OwnerX, OwnerY
			return
		else
			if MyState == EVADE_ST then
				isCircling = false
				return
			end
		end
	end

   -- Patrol when HP and SP are full
	if (OwnerMotion ~= MOTION_STAND) and (OwnerMotion ~= MOTION_SIT) then
		isCircling = false
	end
	if isCircling then
		CircleAroundTarget(OwnerID)
	else
		-- Switch to follow if Owner is on movement or far
		if (not isCloseToOwner()) or (owner_dst == -1) then
			FollowOwner("[IDLE_ST -> FOLLOW_ST] owner is far from me")
			return
		end
	end

end

--------------------------------------------------
function OnFOLLOW_ST()
--------------------------------------------------
	Log("OnFOLLOW_ST") -- comentar

	OwnerMotion = GetV(V_MOTION, OwnerID)
	local CurrTime = GetTick()
	if (OwnerMotion ~= MOTION_STAND) and (OwnerMotion ~= MOTION_SIT) then
		OwnerMoveTime = CurrTime
	end

	-- Owner reached? -> idle
	if isCloseToOwner() then
		if (OwnerMotion == MOTION_STAND) or (OwnerMotion == MOTION_SIT) then -- don't stop until the owner stops
			StopHere("[FOLLOW_ST -> IDLE_ST] Owner reached")
		elseif (OwnerMotion == MOTION_ATTACK) or (OwnerMotion == MOTION_ATTACK2) then
			object = GetV(V_TARGET, OwnerID)
			if (object ~= 0) then
				if object < 50000 then 
					HTact = GetFullTact_LoID(GetMobIDFromActor(object))
				else
					HTact = GetFullTact(GetMobIDFromActor(object))
				end
				if (HTact.Behav ~= BEHA_avoid) then
					MyState = CHASE_ST
					MyEnemy = object
					AtkStartTime = 0
					AtkSkillDoneCount = 0
					vOwnerX, vOwnerY = OwnerX, OwnerY
					Log("[FOLLOW_ST -> CHASE_ST] Cooperative attack")
				end
			end
		elseif (OwnerMotion == MOTION_SKILL) or (OwnerMotion == MOTION_CAST) then   -- only PvM -- configurar Habilitao
			object = GetV(V_TARGET, OwnerID)
			if (object ~= 0) and IsMonster(object) then
				if object < 50000 then 
					HTact = GetFullTact_LoID(GetMobIDFromActor(object))
				else
					HTact = GetFullTact(GetMobIDFromActor(object))
				end
				if (HTact.Behav ~= BEHA_avoid) and (Config[2].ATK_HI_IDS or (object < 100000)) then
					MyState = CHASE_ST
					MyEnemy = object
					AtkStartTime = 0
					AtkSkillDoneCount = 0
					vOwnerX, vOwnerY = OwnerX, OwnerY
					Log("[FOLLOW_ST -> CHASE_ST] Cooperative attack (SKILL/CAST)")
				end
			end
		elseif OwnerMotion == MOTION_HIT then
			if OwnerBlocked == 0 then
				OwnerBlocked = 1
				LstOwnerX    = OwnerX 
                LstOwnerY    = OwnerY
			else
				if (LstOwnerX == OwnerX) and (LstOwnerY == OwnerY) then
					OwnerBlocked = OwnerBlocked + 1
				else
					OwnerBlocked = 0
					LstOwnerX    = 0
					LstOwnerY    = 0
				end
				if OwnerBlocked > 4 then
					StopHere("[FOLLOW_ST -> IDLE_ST] Owner locked.")
					return
				end
			end
			--local angle = math.rad(math.mod(GetTick()/8, 360))
			MyDestX = OwnerX - 1 --math.cos(angle)
			MyDestY = OwnerY --+ math.sin(angle)
			Move(MyID, MyDestX, MyDestY)
		else
			OwnerBlocked = 0
			LstOwnerX    = 0
			LstOwnerY    = 0
			--local angle = math.rad(math.mod(GetTick()/8, 360))
			MyDestX = OwnerX - 1 --math.cos(angle)
			MyDestY = OwnerY --+ math.sin(angle)
			Move(MyID, MyDestX, MyDestY)
		end
		return
	end

	-- Am I bloked?
	if (math.abs(vMercX - MyX) == 0)
	and(math.abs(vMercY - MyY) == 0) then  -- if there is no movement
		if (BlockFound == 0) then
			BlockTime = GetTick() + 300		 -- time-out
			BlockFound = 1
		elseif (BlockFound == 1 and GetTick () > BlockTime) then
			if (OwnerMotion == MOTION_STAND) or (OwnerMotion == MOTION_SIT) then
				local NextTarget = GetMyNextTarget(Self_Per_HP)
				if NextTarget ~= 0 then
					isCircling = false
					MyState = CHASE_ST
					MyEnemy = NextTarget
					vOwnerX, vOwnerY = OwnerX, OwnerY
					Log(string.format("[FOLLOW_ST -> CHASE_ST] Blocked, but target(%d) found", MyEnemy))
					return
				else
					if MyState == EVADE_ST then
						Log("[FOLLOW_ST -> EVADE_ST]")
						return
					end
				end
			end
			BlockFound = 0
			BlockCount = BlockCount + 1
			if BlockCount >= 3 then
				Log("Can't follow in this direction")
				AltAngle = math.rad(math.mod(GetTick(), 360))
				local radius = 12
				MyDestX = MyX + math.cos(AltAngle) * radius
				MyDestY = MyY + math.sin(AltAngle) * radius
				while (radius > 0) and (GetDistance(OwnerX, OwnerY, MyDestX, MyDestY) > Config[Self].TOO_FAR_TARGET) do
					radius = radius - 1
					MyDestX = MyX + math.cos(AltAngle) * radius
					MyDestY = MyY + math.sin(AltAngle) * radius
				end
				if radius > 0 then
					Move(MyID, MyDestX, MyDestY)
					BlockCount = 0
					Log("Can't follow: trying alternate path")
				end
			else
				BlockFound = 0
				MyDestX = OwnerX
				MyDestY = OwnerY
				Move(MyID, OwnerX, OwnerY)
				Log("Can't follow: trying again")
			end
		end
	else
		BlockFound = 0
		BlockCount = 0

		if (CurrTime - OwnerMoveTime > 500) and ((OwnerMotion == MOTION_STAND) or (OwnerMotion == MOTION_SIT)) then
			local NextTarget = GetMyNextTarget(Self_Per_HP)
			
			if NextTarget ~= 0 then
				isCircling = false
				MyState = CHASE_ST
				MyEnemy = NextTarget
				vOwnerX, vOwnerY = OwnerX, OwnerY
				Log(string.format("[FOLLOW_ST -> CHASE_ST] intercepting target(%d)", MyEnemy))
				return
			else
				if MyState == EVADE_ST then
					Log("[FOLLOW_ST -> EVADE_ST]")
					return
				end
			end
		end

		--local angle = math.rad(math.mod(GetTick()/8, 360))
		MyDestX = OwnerX - 1--math.cos(angle)
		MyDestY = OwnerY --+ math.sin(angle)
		Move(MyID, MyDestX, MyDestY)
	end
	vMercX = MyX
	vMercY = MyY

end

--------------------------------------------------
function OnCHASE_ST()
--------------------------------------------------
	-- Log ("OnCHASE_ST")


	if isToFollowAtOnce() then return end

	-- Target is still OK?
	local isBadTarget = false
	local Ex, Ey = GetV(V_POSITION, MyEnemy)
	if (Ex == -1) then
		Log("Target lost")
		isBadTarget = true
	elseif GetDistance(Ex, Ey, OwnerX, OwnerY) >= Config[Self].TOO_FAR_TARGET then
		Log("Target too far")
		isBadTarget = true
	elseif isKS() or not isMobMotionOK(MyEnemy, nil) then
		Log("Target is KS or has suspect motion")
		isBadTarget = true
	end

	-- Get a new target if it's necessary
	
	local NextTarget = GetMyNextTarget(Self_Per_HP)
	if not isBadTarget and NextTarget ~= 0 then
		local EnemyDst = GetDistance2(MyEnemy, MyID)
        if (EnemyDst <= 2) or (EnemyDst <= GetDistance2(NextTarget, MyID)) then
            NextTarget = 0
		end
	end
	if NextTarget ~= 0 then
		MyEnemy = NextTarget
		vOwnerX, vOwnerY = OwnerX, OwnerY
		AtkStartTime = 0
		AtkSkillDoneCount = 0
		Log(string.format("Intercepting new target(%d)", MyEnemy))
	else
		if isBadTarget and MyState ~= EVADE_ST then
			StopHere("[CHASE_ST -> IDLE_ST] no alternative target found")
			return
		end
	end

	-- Homunculus long range attack
	local CurrTime = GetTick()
--	if LRASID ~=0 then
--		if (AtkStartTime == 0) then
--			AtkStartTime = CurrTime
--		end
--		if CanDoAtkSkillsNow() then
--			if DoSkill(LRASID, MyEnemy) then -- ##### AGGRESSIVE SKILL #####
--				AtkSkillDoneCount = AtkSkillDoneCount + 1
--				LastLRAtkTime = CurrTime
--				LastLRAtkID = MyEnemy
--			end
--		end
--	end

	-- Am I in close combat position? -> attack_st
	-- Modificado denovo
	local d = GetV(V_ATTACKRANGE,MyID)
	if (math.abs(Ex - MyX) <= d) and (math.abs(Ey - MyY) <= d) then
		MyState = ATTACK_ST
		if AtkSkillDoneCount == 0 then
			AtkStartTime = 0
		end
		Attack(MyID, MyEnemy)
		Log(string.format("[CHASE_ST -> ATTACK_ST] Target(%d) reached, switch to close combat", MyEnemy))
		return
	end

	-- Modificado para ataque de longo alcance
--	local d = math.floor(math.sqrt((MyX-Ex)^2+(MyY-Ey)^2))
--	if (d ~= -1 and d <= GetV(V_ATTACKRANGE,MyID)) then
--		MyState = ATTACK_ST
--		if AtkSkillDoneCount == 0 then
--			AtkStartTime = 0
--		end
--		Attack(MyID, MyEnemy)
--		Log(string.format("[CHASE_ST -> ATTACK_ST] Target(%d) reached, switch to combat", MyEnemy))
--		return		
--	end

	if Config[Self].LONG_RANGE_SHOOTER == true then
		if isCloseToOwner() then
			Attack(MyID, MyEnemy)
		else
			Log("Pushed away from owner: coming back")
			-- Besides skills that push your homunculus away, there is a Gravity's bug too: even with an empty AI script,
			-- the homunculus moves to arches whom hit him/her
			MyDestX = OwnerX - 1
			MyDestY = OwnerY
			Move(MyID, MyDestX, MyDestY)
		end
	else
		-- Am I bloked?
		if (math.abs(vMercX - MyX) == 0)
		and(math.abs(vMercY - MyY) == 0) then -- if there is no movement
			if (BlockFound == 0) then
				BlockTime = CurrTime + 1000 -- time-out
				BlockFound = 1
			elseif BlockFound == 1 then
				if CurrTime > BlockTime then
					BlockedPathToEnemyID = MyEnemy
					FollowOwner("[CHASE_ST -> FOLLOW_ST] Path to target seems blocked, searching for another target")
					return
				elseif CurrTime > BlockTime - 600 then
					Attack(MyID, MyEnemy) -- it could be a bug position
				end
			end
		else
			BlockFound = 0
			BlockCount = 0
		end

		-- Move to target
		vMercX = MyX; vMercY = MyY
		MyDestX = Ex; MyDestY = Ey
		Move(MyID, MyDestX, MyDestY)
	end
end

--------------------------------------------------
function OnATTACK_ST()
--------------------------------------------------
	-- Log("OnATTACK_ST")
	if MyEnemy == 0 then
		StopHere("[ATTACK_ST -> IDLE_ST] no enemy to attack")
		return
	end	
	
	if isToFollowAtOnce() then return end
	
	local Ex, Ey = GetV(V_POSITION, MyEnemy)
	local posX, posY = GetV(V_POSITION, MyID)
	local d = GetV(V_ATTACKRANGE,MyID)

	-- Get current status information
	local EnemyTarget = GetV(V_TARGET, MyEnemy)

	local CurrTime = GetTick()

	-- Survival Instinct
	if ((EnemyTarget == MyID) and (Self_Per_HP < Config[Self].HP_PERC_DANGER))
	or (HTact.Behav == BEHA_avoid)
	then
		MyState = EVADE_ST
		Evade_ComeBackTime = EVADE_COMEBACK
		Log("[ATTACK_ST -> EVADE_ST] HP < HP_PERC_DANGER or beha=avoid")
		return
	end

	-- Make sure no other monsters are attacking the owner
	if (Config[Self].HELP_OWNER_1ST == true)          and
       (MVP_KS_OK[GetMobIDFromActor(MyEnemy)] == nil) then
		local NewTarget = 0
		NewTarget = GetEnemyOf(OwnerID)
		if NewTarget == MyEnemy then
			NewTarget = 0
		end
		if Config[Self].HELP_FRIEND_1ST then
			for i,v in Friends do
				if MyEnemy == GetEnemyOf(v) then
					NewTarget = 0
				end
			end
		end
		if (NewTarget ~= 0) and (NewTarget ~= MyEnemy) then -- if a threat has been detected
			MyEnemy = NewTarget -- change target
			AtkSkillDoneCount = 0
			AtkStartTime = GetTick() -- a new attack starts (usually this reset is in ON_IDLE_ST)
			
			if NewTarget < 50000 then 
				HTact = GetFullTact_LoID(GetMobIDFromActor(NewTarget))
			else
				HTact = GetFullTact(GetMobIDFromActor(NewTarget))
			end
			
			local Ex, Ey = GetV(V_POSITION, MyEnemy)
			if (math.abs(Ex - MyX) > d) or (math.abs(Ey - MyY) > d) then
				Log("[ATTACK_ST -> CHASE_ST] Switch target -> my owner enemy")
				MyState = CHASE_ST
				if Config[Self].LONG_RANGE_SHOOTER ~= true then
					MyDestX, MyDestY = GetV(V_POSITION, MyEnemy)
					Move(MyID, MyDestX, MyDestY)
				end
				Attack(MyID, MyEnemy)
			else
				Log("Switch target -> my owner enemy (already in close combat position)")
				DoCombat()
			end
			vOwnerX, vOwnerY = OwnerX, OwnerY
			return
		end
	end

	-- Change target when the enemy is dead or lost
	if (Ex == -1) or (MOTION_DEAD == GetV(V_MOTION, MyEnemy)) then
		Log(string.format("MyEnemy(%d) lost or dead. Searching for next target", MyEnemy))
		local NextTarget = GetMyNextTarget(Self_Per_HP)
		if NextTarget ~=0 then
			MyEnemy = NextTarget
			AtkSkillDoneCount = 0
			if isKS() then
				StopHere("[ATTACK_ST -> IDLE_ST]")
				return
			end
			AtkStartTime = GetTick() -- a new attack starts (usually this reset is in ON_IDLE_ST)
		else
			if MyState ~= EVADE_ST then
				StopHere("[ATTACK_ST -> IDLE_ST] No more immediate targets")
			end
			return
		end
	end
	
	-- Intercept the enemy when he/she moves
	if Config[Self].LONG_RANGE_SHOOTER ~= true then
		if (math.abs(Ex - MyX) > d) or (math.abs(Ey - MyY) > d) then
			Log(string.format("[ATTACK_ST -> CHASE_ST] Enemy(%d) moved to (%d, %d), while I was at (%d, %d)", MyEnemy, Ex, Ey, MyX, MyY))
			MyState = CHASE_ST
			MyDestX = Ex; MyDestY = Ey
			Move(MyID, MyDestX, MyDestY)
			Attack(MyID, MyEnemy)
			return
		end
	end

	local Ex, Ey = GetV(V_POSITION, MyEnemy)
	local posX, posY = GetV(V_POSITION, MyID)
	
	if (math.abs(Ex - posX) <= d) and (math.abs(Ey - posY) <= d) then
		DoCombat()
		posX, posY = MercDancing(posX, posY)
		if (MercType < 11) and (math.abs(Ex - posX) < Config[Self].MIN_DIST) 
		                   and (math.abs(Ey - posY) < Config[Self].MIN_DIST) then
			if math.abs(Ex - posX) < Config[Self].MIN_DIST then
				if posX >= Ex then posX = posX + 1 end
				if posX < Ex  then posX = posX - 1 end
			end
			if math.abs(Ey - posY) < Config[Self].MIN_DIST then
				if posY >= Ey then posY = posY + 1 end
				if posY < Ey  then posY = posY - 1 end
			end
			if GetDistance(OwnerX, OwnerY, posX, posY) > Config[Self].TOO_FAR_TARGET - 2 then
				posX, posY = OwnerX, OwnerY
			end
		end		
		Move(MyID, posX, posY)	
		IsFevDanc = false
	end	
	
	local Danc_Count = 0	
	while (Config[Self].CONFG_VAR_DANC == 2) and (Danc_Count < 1) do	
		posX, posY = GetV(V_POSITION, MyID)
		Ex, Ey = GetV(V_POSITION, MyEnemy)
		if (Ex == -1) or (Ey == -1) or (MOTION_DEAD == GetV(V_MOTION, MyEnemy)) then
			break
		end
		if (math.abs(Ex - posX) <= d) and (math.abs(Ey - posY) <= d) then
			DoCombat()
			posX, posY = MercDancing(posX, posY)
			if (MercType < 11) and (math.abs(Ex - posX) < Config[Self].MIN_DIST) 
							and (math.abs(Ey - posY) < Config[Self].MIN_DIST) then
				if math.abs(Ex - posX) < Config[Self].MIN_DIST then
					if posX >= Ex then posX = posX + 1 end
					if posX < Ex  then posX = posX - 1 end
				end
				if math.abs(Ey - posY) < Config[Self].MIN_DIST then
					if posY >= Ey then posY = posY + 1 end
					if posY < Ey  then posY = posY - 1 end
				end
				if GetDistance(OwnerX, OwnerY, posX, posY) > Config[Self].TOO_FAR_TARGET - 2 then
					posX, posY = OwnerX, OwnerY
				end
			end		
			Move(MyID, posX, posY)	
			IsFevDanc = false
		end
		Danc_Count = Danc_Count + 1
	end
	
	-- bug position check 
--[[
	if (CurrTime - AtkStartTime > 30000) -- after 30"
	and (GetV(V_MOTION, MyID) == MOTION_STAND) -- this don't seem actually to work [...]
	then
		if BugPosiTime == 0 then
			BugPosiTime = CurrTime
			Log("Bug position check has been started")
		end
		-- Log(string.format("Bug position check: my motion = %d", GetV(V_MOTION, MyID)))
		if CurrTime - BugPosiTime > 1000 then
			MyState = BUGPOSI_ST
			AtkStartTime = GetTick()
			AtkSkillDoneCount = 0
			Log("[ATTACK_ST -> BUGPOSI_ST] I'm doing no attack movement (Move=%d): starting anti bug position motion")
			return
		end
	else
		BugPosiTime = 0
	end ]]--

	return
end

--------------------------------------------------
function OnEVADE_ST()
--------------------------------------------------
	Log ("OnEVADE_ST")

	-------- evasion-end conditions ---------------

	-- 1: if the enemy changes target
	local EnemyTarget = GetV(V_TARGET, MyEnemy)
	if (EnemyTarget ~= MyID) then
		StopHere("[EVADE_ST -> IDLE] Enemy no longer aims at me")
		return
	end

	-- 2: if the enemy is out of sight
	if (true == IsOutOfSight(MyID, MyEnemy)) then
		StopHere("[EVADE_ST -> IDLE] Enemy out of sight")
		return
	end

	-- 3: if the enemy is dead
	if (MOTION_DEAD == GetV(V_MOTION, MyEnemy)) then
		StopHere("[EVADE_ST -> IDLE] Enemy dead")
		return
	end

	-- 4: if HP are OK now
	if (Self_Per_HP > Config[Self].HP_PERC_DANGER) then
		if (HTact.Behav ~= BEHA_avoid) and (HTact.Behav ~= BEHA_coward) then
			MyState = ATTACK_ST
			AtkStartTime = 0
			AtkSkillDoneCount = 0
			vOwnerX, vOwnerY = OwnerX, OwnerY
		 	Log("[EVADE_ST -> ATTACK] HPs are ok now")
		  	return
		end
	end
	
	------- Evading skills ------------------------
	for i,v in SkillSlot do
		if (SkillSlot[i].Type == 2) or ((SkillSlot[i].Type == 6)) then
			DoSkill(i, MyEnemy)			
		end
		if (SkillSlot[i].Type == 4) or (SkillSlot[i].Type == 5) then
			DoSkill(i, MyID)
		end
	end

	------- Evading maneuvres ---------------------
	CircleAroundTarget(OwnerID)
	return
end

--------------------------------------------------
function OnBUGPOSI_ST()
--------------------------------------------------
	local CurrTime = GetTick()
	
	MoveToOwner(MyID)
	if (CurrTime - AtkStartTime > 1500) or (GetDistance2(OwnerID, MyID) <= 1) then
		StopHere("[BUGPOSI_ST -> IDLE_ST] Bugposi movement is over")
	end
end

--------------------------------------------------
-- ############ OUTPUT MOVEMENTS ##########
--------------------------------------------------

--------------------------------------------------
function OnOUTMOVE_NEWFRIEND_ST()
--------------------------------------------------
	local CurrTime = GetTick() - OutMoveTime
	if CurrTime > 3000 then
		FollowOwner("[ -> FOLLOW_ST] Output movement done")
		return
	else
		local OutMoveX, OutMoveY = GetV(V_POSITION, OutMoveTarget)
		local angle = math.rad(math.mod(CurrTime/4, 360))
		local X = OutMoveX + math.cos(angle) * 3
		local Y = OutMoveY + math.sin(angle) * 3
		Move(MyID, X, Y)
		MyDestX = X
		MyDestY = Y
	end
end

--------------------------------------------------
function OnOUTMOVE_DELFRIEND_ST()
--------------------------------------------------
	local CurrTime = GetTick() - OutMoveTime
	if CurrTime > 3000 then
		FollowOwner("[ -> FOLLOW_ST] Output movement done")
		return
	else
		local OutMoveX, OutMoveY = GetV(V_POSITION, OutMoveTarget)
		local angle = math.rad(math.mod(CurrTime/4, 360))
		local X = OutMoveX + math.cos(angle) * 3
		local Y = OutMoveY
		Move(MyID, X, Y)
		MyDestX = X
		MyDestY = Y
	end
end

--------------------------------------------------
-- ############# COMMAND STATUS ###########
--------------------------------------------------

--------------------------------------------------
function OnMOVE_CMD_ST()
--------------------------------------------------
	Log("OnMOVE_CMD_ST")

	if (MyX == MyDestX and MyY == MyDestY) then
		StopHere("destination reached")
	else
		Move(MyID, MyDestX, MyDestY)
	end
end

--------------------------------------------------
function OnSTOP_CMD_ST()
--------------------------------------------------
end

--------------------------------------------------
function OnATTACK_OBJECT_CMD_ST()
--------------------------------------------------
end

--------------------------------------------------
function OnATTACK_AREA_CMD_ST()
--------------------------------------------------
	Log ("OnATTACK_AREA_CMD_ST")

	local	object = GetEnemyOf(OwnerID)
	if (object == 0) then
		object = GetMyEnemy_AnyNoKS(MyID)
	end

	if (object ~= 0) then -- MYOWNER_ATTACKED_IN or ATTACKED_IN
		MyState = CHASE_ST
		MyEnemy = object
		vOwnerX, vOwnerY = OwnerX, OwnerY
		return
	end

	if (MyX == MyDestX and MyY == MyDestY) then -- DESTARRIVED_IN
		MyState = IDLE_ST
		IdleStartTime = GetTick()
		MyEnemy = 0
	end

end

--------------------------------------------------
function OnPATROL_CMD_ST()
--------------------------------------------------
	Log("OnPATROL_CMD_ST")

	local	object = GetEnemyOf(OwnerID)
	if (object == 0) then
		object = GetMyEnemy_AnyNoKS(MyID)
	end

	if (object ~= 0) then -- MYOWNER_ATTACKED_IN or ATTACKED_IN
		MyState = CHASE_ST
		MyEnemy = object
		vOwnerX, vOwnerY = OwnerX, OwnerY
		Log("[PATROL_CMD_ST -> CHASE_ST] ATTACKED_IN")
		return
	end

	if (MyX == MyDestX and MyY == MyDestY) then -- DESTARRIVED_IN
		MyDestX = MyPatrolX
		MyDestY = MyPatrolY
		MyPatrolX = x
		MyPatrolY = y
	end
   Move(MyID,MyDestX,MyDestY)

end

--------------------------------------------------
function OnHOLD_CMD_ST ()
--------------------------------------------------
	Log("OnHOLD_CMD_ST")

	if (MyEnemy ~= 0) then
		local d = GetDistance(MyEnemy,MyID)
		if (d ~= -1 and d <= GetV(V_ATTACKRANGE,MyID)) then
			Attack (MyID,MyEnemy)
		else
			MyEnemy = 0
		end
		return
	end

	local	object = GetEnemyOf(OwnerID)
	if (object == 0) then
		object = GetMyEnemy_AnyNoKS(MyID)
		if (object == 0) then
			return
		end
	end

	MyEnemy = object

end

--------------------------------------------------
function OnSKILL_OBJECT_CMD_ST()
--------------------------------------------------
end

--------------------------------------------------
function OnSKILL_AREA_CMD_ST()
--------------------------------------------------
	Log("OnSKILL_AREA_CMD_ST")

	if (GetDistance(MyX, MyY, MyDestX, MyDestY) <= GetV(V_SKILLATTACKRANGE, MyID, MySkill)) then -- DESTARRIVED_IN
		SkillGround (MyID, MySkillLevel, MySkill, MyDestX, MyDestY)
		MyState = IDLE_ST
		IdleStartTime = GetTick()
		MySkill = 0
	end

end

--------------------------------------------------
function OnFOLLOW_CMD_ST ()
--------------------------------------------------
	Log("OnFOLLOW_CMD_ST")

	OwnerMotion = GetV(V_MOTION, OwnerID)
	if not isCloseToOwner()	and (OwnerMotion ~= MOTION_STAND) and (OwnerMotion ~= MOTION_SIT)
	then
		--local angle = math.rad(math.mod(GetTick()/8, 360))
		MyDestX = OwnerX - 2--math.cos(angle)
		MyDestY = OwnerY --+ math.sin(angle)
		Move(MyID, MyDestX, MyDestY)
	end	
end

--------------------------------------------------
-- ############# MORE FUNCTIONS ###########
--------------------------------------------------

--------------------------------------------------
function GetMyNextTarget(MercHPPerc)
--------------------------------------------------
	local object = 0

	-- 1: SAFETY FIRSTY! Is the homunculus is under attack?
	local Agressor = GetMyEnemy_AttackingMe(MyID) -- get TactData too
	local AgBehav, AgSkill, AgLevel, AgAlche
	
	if Config[Self].HELP_FRIEND_1ST then
		for i,v in Friends do
			object = GetEnemyOf(v)
			if GetV(V_MOTION, object) == MOTION_DEAD then
				object = 0
			end
			if object ~= 0 then
				break
			end
		end
	end
	
	if Agressor ~= 0 then
		if (MercHPPerc < Config[Self].HP_PERC_DANGER) or (HTact.Behav == BEHA_avoid) or (HTact.Behav == BEHA_coward) then
			MyState = EVADE_ST -- flee away if HP < secure level
			MyEnemy = Agressor
			Evade_ComeBackTime = EVADE_COMEBACK
			Log(string.format("[ -> EVADE_ST] GetMyNextTarget() attacked by enemy(%d) (avoid/coward mode or while HP < HP_PERC_DANGER)", MyEnemy))
			return 0
		else -- we are not in danger for now: save these data and let's check for other threats first
         AgBehav = HTact.Behav
         AgSkill = HTact.Skill
         AgLevel = HTact.Level
         AgAlche = HTact.Alche
		 
		if Config[Self].KILL_YOUR_ENEMIES_1ST and (object == 0) then -- ... unless KILL_YOUR_ENEMIES_1ST is enabled
				Log(string.format("GetMyNextTarget() <defensive scan + KILL_YOUR_ENEMIES_1ST> attacked by enemy(%d)", Agressor))
		      HTact.Behav = AgBehav
		      HTact.Skill = AgSkill
		      HTact.Level = AgLevel
		      HTact.Alche = AgAlche
				return Agressor
			end
		end
	end

	-- 2: if the owner or a friend is under attack, help him/her
	if object == 0 then
		object = GetEnemyOf(OwnerID)
		if object ~= 0 then
			if GetV(V_MOTION, object) == MOTION_DEAD then
				object = 0
			end
		end
	end

	if object ~= 0 then
		Log(string.format("GetMyNextTarget() the owner or a friend is attacked by enemy(%d):", object))
		
		if object < 50000 then 
			HTact = GetFullTact_LoID(GetMobIDFromActor(object))
		else
			HTact = GetFullTact(GetMobIDFromActor(object))
		end
		if (HTact.Behav ~= BEHA_avoid) then
			return object
		end
	end

	-- 3: if the homunculus is under attack (and his/her life is not in danger, see priority 1)
	if Agressor ~= 0 then
		Log(string.format("GetMyNextTarget() <defensive scan> attacked by enemy(%d)", Agressor))
		HTact.Behav = AgBehav
		HTact.Skill = AgSkill
  		HTact.Level = AgLevel
    	HTact.Alche = AgAlche
		return Agressor
	end

	for i,v in Friends do
		object = GetEnemyOf(v)
		if GetV(V_MOTION, object) == MOTION_DEAD then
			object = 0
		end
		if object ~= 0 then
			break
		end
	end
	
	if object ~= 0 then
		Log(string.format("GetMyNextTarget() a friend is attacked by enemy(%d):", object))
		
		if object < 50000 then 
			HTact = GetFullTact_LoID(GetMobIDFromActor(object))
		else
			HTact = GetFullTact(GetMobIDFromActor(object))
		end
		if (HTact.Behav ~= BEHA_avoid) then
			return object
		end
	end	

	-- 4: if the owner or a friend is attacking --> cooperate
	local OwnerMotion = GetV(V_MOTION, OwnerID)
	local CooTarget = 0
	object = GetV(V_TARGET, OwnerID)
	if (object ~= 0) then
		if OwnerMotion == MOTION_ATTACK or OwnerMotion == MOTION_ATTACK2 then
			CooTarget = object
		elseif OwnerMotion == MOTION_CAST or OwnerMotion == MOTION_SKILL then
			if IsMonster(object) and (Config[2].ATK_HI_IDS or (object < 100000)) then
				CooTarget = object
			end
		end
	end
	if CooTarget == 0 then -- Owner is not attacking. Maybe a friend?
		local FriendMotion = 0
		for i,v in Friends do
			object = GetV(V_TARGET, v)
			FriendMotion = GetV(V_MOTION, v)
			if (object ~= 0) then
				if FriendMotion == MOTION_ATTACK or FriendMotion == MOTION_ATTACK2 then
					CooTarget = object
					break
				elseif FriendMotion == MOTION_CAST or FriendMotion == MOTION_SKILL then
					if IsMonster(object) and (Config[2].ATK_HI_IDS or (object < 100000)) then
						CooTarget = object
						break
					end
				end
			end
		end
	end
	if CooTarget ~= 0 then
		Log(string.format("GetMyNextTarget() cooperative target(%d) found:", CooTarget))
		if CooTarget < 50000 then 
			HTact = GetFullTact_LoID(GetMobIDFromActor(CooTarget))
		else
			HTact = GetFullTact(GetMobIDFromActor(CooTarget))
		end
		if (HTact.Behav ~= BEHA_avoid) then
			return CooTarget
		end
	end

	-- 5: aggressive, if HPs are OK and a target is near
	if (MercHPPerc > Config[Self].HP_PERC_SAFE2ATK) and (GetV(V_SP, MyID) >= Config[Self].SP_SAFE2ATK)
	then
		object = GetMyEnemy_AnyNoKS(MyID) -- get TactData too
		if (object ~= 0) then
			Log(string.format("GetMyNextTarget() <aggressive scan> target(%d) found", object))
			return object
		end
	end

	return 0
end

--------------------------------------------------
function isToFollowAtOnce()
-- used in OnCHASE_ST() and in OnATTACK_ST()
--------------------------------------------------
	if GetV(V_MOTION, OwnerID) == MOTION_MOVE then
		if Config[Self].FOLLOW_AT_ONCE > 0 then
			local MyEnX, MyEnY = GetV(V_POSITION, MyEnemy)
			local ExtraDst = 0
			if GetDistance(OwnerX, OwnerY, MyX, MyY) < 4 then ExtraDst = 1 end
			if GetDistance(OwnerX, OwnerY, MyEnX, MyEnY) > GetDistance(vOwnerX,vOwnerY, MyEnX, MyEnY) + ExtraDst then
				FollowOwner("[ -> FOLLOW_ST] Owner is moving away")
				return true
			end
		else -- don't follow at once, keep the homunculus in range
			if GetDistance(OwnerX, OwnerY, MyX, MyY) > Config[Self].TOO_FAR_TARGET - 3 then
				FollowOwner("[ -> FOLLOW_ST] Owner is moving away")
				return true
			end
		end
		vOwnerX = OwnerX; vOwnerY = OwnerY
	end
	return false
end

--------------------------------------------------
function isMobMotionOK(MobID, PlayerList)
-- before to engage in combat with this monster (that seems free because he/she targets nobody and nobody seems to target
-- her/him) check his/her motion. N.B.: the "Get" functions provided by Gravity are very limited, so we can do little for
-- this and the homunculus in many cases can only try to guess what is happening. I hope Gravity will improve/extend them
--------------------------------------------------
	if 0 == IsMonster(MobID) then return true; end; -- for PvP mods

	local CurrTime = GetTick()
	local mob_motion = GetV(V_MOTION, MobID)
	local mob_type = GetMobIDFromActor(MobID)

	local mob_dist = 100
	local no_players = true
	if Config[Self].IGNORE_AMC and (Config[Self].NO_MOVING_TARGETS or Config[Self].ADV_MOTION_CHECK) then
		mob_dist = GetDistanceFromOwner(MobID)
		if PlayerList ~= nil then
			for k,v in pairs(PlayerList) do
				if IsMonster(v) == 0 then
					if (v ~= MyID) and (v ~= OwnerID) and (Friends[v] == nil) then
						no_players = false
						break
					end
				end
			end
		else
			no_players = true
		end
	end
	
--MOTION_MOVE: is he/she following another player?
	if mob_motion == MOTION_MOVE then
		if Config[Self].NO_MOVING_TARGETS == true then
			if Config[Self].IGNORE_AMC and no_players and mob_dist <= Config[Self].IGNORE_AMC_RANGE then
				return true
			end
			isCircling = false
			Log("---mob is moving")
			return false -- This option is good also to wait until the monsters come close to you
		else
			return true -- [...] (to do: find a way to determine if the moster is following someone in order to make the AI
			-- more kind and friendly) N.B.: tecnically it's not KS to intercept a monster before someone else (the combat
			-- is not yet started and who "mob" just make longer this phase), but it looks quite unfriendly to many players.
		end
--MOTION_STAND: is he/she trapped or frozen?
	elseif mob_motion == MOTION_STAND then
		if Config[Self].ADV_MOTION_CHECK == true then
			-- this is optional, because it's not reliable and gives false positives
			if Config[Self].IGNORE_AMC and no_players and mob_dist <= Config[Self].IGNORE_AMC_RANGE then
				return true
			end
			if NoAggroMob[mob_type] == nil then -- is this an immobile or passive monster?
				Log("---mob trapped")
				return false -- mobile/aggro mob = trapped, frozen or obstacled (exceptions: [...] temporary passive mode, archers)
			else
				return true -- immobile/passive mob. [...] (to do: find a way to chek if it's trapped) that's tricky: passive
				-- monsters stands at regular intervals and immobile mobs stands all the time, but even them could get trapped
				-- (e.g. Clocks) or frozen (e.g. Zenorks); also archers are usually immobile. Maybe we could check the distance from
				-- other players, but for example Zenork are looters and so distance has not great meaning.
			end
		else
			return true
		end
-- MOTION_HIT: who hit him/her?
	elseif mob_motion == MOTION_HIT and MyState ~= ATTACK_ST then
		if Config[Self].ADV_MOTION_CHECK == true then
			-- this is optional, because it's not reliable and gives false positives
			if (CurrTime - LastLRAtkTime < 2000) and (LastLRAtkID == MobID) then
				return true -- happened shortly after our long ranged attack at this monster: it's our hit. But [...] it could
				-- happen that someone hits our target in the meantime.
			else
				if Config[Self].IGNORE_AMC and no_players and mob_dist <= Config[Self].IGNORE_AMC_RANGE then
					return true
				end
				Log("---mob hit by area atk")
				return false -- it seems an hit from another player, maybe an area attack
			end
		else
			return true
		end
-- MOTION_DEAD
	elseif mob_motion == MOTION_DEAD then
		Log("---mob dead")
		return false
	end

-- MOTION_ATTACK, MOTION_ATTACK2, MOTION_SKILL, MOTION_CAST should require some targetings (unless they are area attacks) and so detected before
-- MOTION_PICKUP and MOTION_SIT shouldn't be in monsters

	return true
end

--------------------------------------------------
function isKS()
-- used in OnCHASE_ST()
--------------------------------------------------
	if 0 == IsMonster(MyEnemy) then return false; end -- for PvP mods

---------------------------------------------------------------------------------------
	local MyEnemyMobID = 0

	if I_AM_HOMUN then
		MyEnemyMobID = GetV(V_HOMUNTYPE, MyEnemy)
	else
		MyEnemyMobID = GetMobIDFromActor(MyEnemy)
	end
	
	if MVP_KS_OK[MyEnemyMobID] ~= nil then
--		Log("MVP atacando. Verificao de KS desabilitada.")
		return false
	end
	
	if MyEnemy == CanDoKS then return false end
----------------------------------------------------------------------------------------

	-- look at monster target
	local mob_target = GetV(V_TARGET, MyEnemy)
	if mob_target > 0 then
		--- FUNCAO MODIFICADA
		local bool_temp = false
		if (0 ~= IsMonster(mob_target)) then
			bool_temp = true
		elseif (mob_target < 100000) then
			if (GetV(V_POSITION, mob_target) == -1) then
				bool_temp = true
			end
		end	
		if (mob_target == MyID) or (mob_target == OwnerID) or
		   (Friends[mob_target] ~= nil) or bool_temp then
			return false
		else
			Log("MyEnemy was already in battle, action aborted (anti-KS)")
			return true
		end
	end

	-- look at (not friend) player targets
	for i,v in ipairs(Scre_Actors) do
		if (v ~= OwnerID) and (v ~= MyID) then
			if 0 == IsMonster(v) then -- if it is a player
				if isNotFriend(v) then -- and not a friend
					if GetV(V_TARGET, v) == MyEnemy then
						Log("Another player is aiming at MyEnemy, action aborted (anti-KS)")
						return true
					end
				end
			end
		end
	end

	return false
end

--------------------------------------------------
function StopHere(s)
--------------------------------------------------
	MyState = IDLE_ST
	MyEnemy = 0
	MySkill = 0
	IdleStartTime = GetTick()
	if Config[Self].LONG_RANGE_SHOOTER ~= true and not (MyDestX == OwnerX + 1 and MyDestY == OwnerY) then
		MyDestX, MyDestY = MyX, MyY
		Move(MyID, MyDestX, MyDestY) -- stop at once
	else
		MyDestX, MyDestY = MyX, MyY
	end
	Log("#STOP#" .. s)
end

--------------------------------------------------
function FollowOwner(s)
--------------------------------------------------
	MyState = FOLLOW_ST
	MyEnemy = 0
	MySkill = 0
	BlockFound = 0
	MyDestX, MyDestY = OwnerX, OwnerY
	MoveToOwner(MyID)
	Log(s)
end

--------------------------------------------------
function DoCombat()
--------------------------------------------------
	if (AtkStartTime == 0) then
		AtkStartTime = GetTick()
	end
	
	if (MyEnemy == MyID) or (MyEnemy == OwnerID) or (Friends[MyEnemy] ~= nil) then
		Log(MyEnemy.." no  um alvo valido para combate.")
		return
	end
	
	if (MySkill == 0) then
--		Debuff
		local EnemyType = GetMobIDFromActor(MyEnemy)
		Attack(MyID, MyEnemy)
		if (EnemyType < 1078 or EnemyType > 1085) then -- don't waste SP on plants and mushrooms
			-- Attack e buff skills
			for i,v in SkillSlot do				
				if (SkillSlot[i].Type == 6) and (SkillSlot[i].Level ~= 0) then
					if CanDoAtkSkillsNow() then
						DoSkill(i, MyEnemy)
					end 
				end
				if ((SkillSlot[i].Type == 1) or (SkillSlot[i].Type == 2)) and (SkillSlot[i].Level ~= 0) then
					if CanDoAtkSkillsNow() then
						if DoSkill(i, MyEnemy) then
							AtkSkillDoneCount = AtkSkillDoneCount + 1
						end
					end 
				end -- SkillEn[8214]
				if SkillSlot[i].SkillID == 8214 then -- Disparo Violento
					if GetDistance2(MyID, MyEnemy) <= Config[Self].CA_MIN_DIST then 
						DoSkill(i, MyEnemy)
					end 
				end
				if (SkillSlot[i].Type == 3) and (SkillSlot[i].Level ~= 0) then
					if CanDoAtkSkillsNow() then
						if SkillSlot[i].SkillID == 8217 then
							if ExecBrandish(i, MyEnemy, 7) then
								AtkSkillDoneCount = AtkSkillDoneCount + 1
							end
						elseif CountEnemiesClose(MyEnemy, SkillSlot[i].Range) >= Config[Self].NR_M_EN_AREA then
							if DoSkill(i, MyEnemy) then
								AtkSkillDoneCount = AtkSkillDoneCount + 1
							end
						end
					end 
				end
				if SkillSlot[i].Type == 5 then
					DoSkill(i, MyID)
				end
			end
		end
	end

	if (MySkill ~= 0) then
		SkillObject(MyID, MySkillLevel, MySkill, MyEnemy)
		MySkill = 0
	end
end

--------------------------------------------------
function CanDoAtkSkillsNow()
-- is it the right time for an aggressive skill?
--------------------------------------------------
	local result = true
	local CurrTime = GetTick()
	local TimeOutElapsed = false
	if (CurrTime - AtkStartTime) > Config[Self].SKILL_TIME_OUT then
		TimeOutElapsed = true
	end
	
	if (HTact.Skill == WITH_no_skill) then
		result = false
	elseif HTact.Skill == WITH_one_skill then
		-- Log(string.format("AtkSkillDoneCount = %d, CurrTime - AtkStartTime = %d", AtkSkillDoneCount, CurrTime - AtkStartTime))
		if (AtkSkillDoneCount > 1) or  TimeOutElapsed then
			result = false
		end
	elseif HTact.Skill == WITH_two_skills then
		if (AtkSkillDoneCount > 2) then
			result = false
		end
	elseif HTact.Skill == WITH_max_skills then
		if TimeOutElapsed then
			result = false
		end
	elseif HTact.Skill == WITH_slow_power then
		if (Config[Self].LONG_RANGE_SHOOTER ~= true) and (CurrTime - AtkStartTime < DELAY_SLOW_POWER) then
			result = false
		end
	end
	return result
end

--------------------------------------------------
function DoSkill(s_index, Target)
--------------------------------------------------
	Log("DoSkill "..SkillSlot[s_index].SkillID.." on "..Target..".")
	if SkillSlot[s_index].Level == 0 then
		return false
	end
	local CurrTime = GetTick()
	if CastDelayEnd > CurrTime then
		return false
	end
	
	local result = false
	
	if SkillSlot[s_index].Engaged then -- if the skill is already active (or in delay time), wait until it goes OFF
		if CurrTime > SkillSlot[s_index].TimeOut then
			SkillSlot[s_index].Engaged = false
		end
	else -- if the skill is OFF, activate it
		
		if SkillSlot[s_index].LastEnemy == Target and SkillSlot[s_index].Type == 6 then
			return false
		end
		if Self_SP >= SkillSlot[s_index].MinSP then -- if there are enough SP left
			MySkill = SkillSlot[s_index].SkillID
			MySkillLevel = SkillSlot[s_index].Level
			SkillSlot[s_index].TimeOut = CurrTime + SkillSlot[s_index].HowLast
			SkillSlot[s_index].Engaged = true
			SkillSlot[s_index].LastEnemy = Target
			SkillObject(MyID, MySkillLevel, MySkill, Target)
			Log(string.format("Done skill %d lvl %d on target %d", MySkill, MySkillLevel, Target))
			if Config.CONFG_VAR_USE_VDLY then
				
				AddDelaySkill = (Mast_HP*Self_HP*Self_Max_SP^2*250)/(Mast_Max_HP*Self_Max_HP*Self_SP^2)
				
				if AddDelaySkill <  100 then AddDelaySkill = 100 end
				Log("AddDelaySkill set to "..AddDelaySkill)
			elseif (HTact.Skill == WITH_full_power) then
				AddDelaySkill = 100
			else
				AddDelaySkill = 1000
			end
			CastDelayEnd = CurrTime + AddDelaySkill
			result = true
			MySkill = 0
		end
	end
	return result
end

--------------------------------------------------
function CircleAroundTarget(TrgID)
--------------------------------------------------
	Log("### CIRCLE START ###")
	-- Log(string.format("Step move=%d, Evade_ComeBackTime=%d, CircleBlockStep=%d", CircleDir, Evade_ComeBackTime, CircleBlockStep))

	local TrgX, TrgY		= GetV(V_POSITION, TrgID)

	--------------- current destination -----------
	if (Evade_ComeBackTime == EVADE_COMEBACK) or (CircleBlockStep == 2) then
		local angle = math.rad(math.mod(GetTick()/8, 360))
		MyDestX = OwnerX - math.cos(angle)
		MyDestY = OwnerY - math.sin(angle)
	else
		MyDestX = TrgX - AAI_CIRC_X[CircleDir] -- direction defined by the circle step
		MyDestY = TrgY - AAI_CIRC_Y[CircleDir]
	end

	--------------- obstacle detection ------------
	if (math.abs(vMercX - MyX) == 0)
	and(math.abs(vMercY - MyY) == 0) then -- if there is no movement
		if (CircleBlockStep == 0) then
			CircleBlockTime = GetTick() + 700 -- time-out
			CircleBlockStep = 1
		elseif (CircleBlockStep == 1) then -- it is in wait mode until the time out expires
			if (GetTick () > CircleBlockTime) then -- if it expires
				CircleBlockStep = 2 -- we assume that an obstacle was found in the current direction
			end
		end
	else
		CircleBlockStep = 0
	end

	--------------- position reached? -------------
	if  (math.abs(MyDestX - MyX) < 1)
	and (math.abs(MyDestY - MyY) < 1) -- yes!
	then									  -- select next direction
		if (MyState == EVADE_ST) then
			if CircleBlockStep == 2 then
				CircleDir = CircleDir - 1
				Evade_ComeBackTime = 0
			else
				if (Evade_ComeBackTime == EVADE_COMEBACK) then
					if (GetDistance2(MyEnemy, OwnerID) <=1) then
						CircleDir = CircleDir - 2 -- the monster shoud reach the owner, so he/she can help
						Evade_ComeBackTime = 0
					end
				else
					CircleDir = CircleDir - 2
					Evade_ComeBackTime = Evade_ComeBackTime + 1
				end
			end
		else
			CircleDir = CircleDir - 1
			Evade_ComeBackTime = 0
		end
		if (CircleDir < 1) then
			CircleDir = AAI_CIRC_MAXSTEP
		end
	end

	----- record curren position and move on ------
	vMercX = MyX
	vMercY = MyY
	Move(MyID, MyDestX, MyDestY)
	Log("=== CIRCLE END ===")
	return
end

--------------------------------------------------
function GetEnemyOf(id) -- this function is not for PvP Mods
--------------------------------------------------
	local result = 0
	local actors = GetActors()
	local enemies = {}
	local enemies2 = {}
	local index = 1
	local index2 = 1
	local target
	for i,v in ipairs(actors) do
		if (v ~= OwnerID and v ~= myid) then
			if (IsMonster(v) == 1) then
				if GetV(V_MOTION, v) ~= MOTION_DEAD then
					target = GetV(V_TARGET, v)
					if (target == id) then
						enemies[index] = v
						index = index + 1
					end
				end
			end
		end
	end

	local min_dis = 100
	local dis
	
	for i,v in ipairs(enemies) do
		dis = GetDistance2(MyID, v)
		if (dis < min_dis) then
			result = v
			min_dis = dis
		end
	end
	
	if Config[2].IDS_AT_ORD == 1 or Config[2].IDS_AT_ORD == 2 then
		for i,v in ipairs(enemies) do
			dis = GetDistance2(MyID, v)
--			Log("enemies["..i.."] = "..v.." dis "..dis)
			if (dis <= min_dis + 3) then
				enemies2[index2] = v
				index2 = index2 + 1
			end
		end
		for i,v in ipairs(enemies2) do
--			Log("enemies2["..i.."] = "..v)
			if ((Config[2].IDS_AT_ORD == 1) and (v < result)) or
			   ((Config[2].IDS_AT_ORD == 2) and (v > result)) then
				result = v
			end
		end
	end
	
--	if result ~= 0 then Log("GetEnemyOf("..id.."): "..(index-1).." monstros atacando. Selecionado: "..result) end
	return result
end

--------------------------------------------------
function CountEnemiesClose(target, range)
--------------------------------------------------
	local ecount = 0
	local actors = GetActors()
	
	local TrgX, TrgY = GetV(V_POSITION, target)
	if TrgX == -1 then return 0 end
	
	local TempX, TempY
	
	for i,v in ipairs(actors) do
		if 1 == IsMonster(v) then
			TempX, TempY = GetV(V_POSITION, v)
			if (TempX ~= -1)                              and
			   (math.abs(TrgX - TempX) <= (range - 1)/ 2) and
			   (math.abs(TrgY - TempY) <= (range - 1)/ 2) and
			   (MOTION_DEAD ~= GetV(V_MOTION, v))         then 
				ecount = ecount + 1
			end
		end
	end
	return ecount
end

--------------------------------------------------
function GetMyEnemy_AttackingMe(myid)
--------------------------------------------------
	-- Log(string.format("---- DEFENSIVE SCAN STARTED for ID (%d)", myid))
	local result = 0
	local actors = GetActors()
	local enemy_1st = {} -- primary targets
	local index_1st = 1
	local enemy_std = {} -- normal targets
	local index_std = 1
	local enemy_lss = {} -- last targets
	local index_lss = 1
	local enemy_wkk = {} -- weak targets
	local index_wkk = 1
	local mob_type, mob_target, mobBehav, ListedType

	-- Fill a list of enemies who have me as target
	for i,v in ipairs(actors) do
		if GetV(V_MOTION, v) ~= MOTION_DEAD then
			if (v ~= OwnerID) and (v ~= MyID) then
				mob_target = GetV(V_TARGET, v)
				if mob_target == MyID then
					mob_type = GetMobIDFromActor(v) -- get tact data
					if v < 50000 then
						mobBehav = GetTact_LoID(V_TACTBEHAV, mob_type)
					else
						mobBehav = GetTact(V_TACTBEHAV, mob_type)
					end
					if    (mobBehav == BEHA_avoid)
					or    (mobBehav == BEHA_coward) then
						result = v
						break -- this monster is too dangerous: evade from him/her at once!
					elseif (mobBehav == BEHA_attack_1st) or (mobBehav == BEHA_react_1st) then
						enemy_1st[index_1st] = v
						index_1st = index_1st + 1
					elseif (mobBehav == BEHA_attack_last) or (mobBehav == BEHA_react_last) then
						enemy_lss[index_lss] = v
						index_lss = index_lss + 1
					elseif mobBehav == BEHA_attack_weak then
						enemy_wkk[index_wkk] = v
						index_wkk = index_wkk + 1
					else
						enemy_std[index_std] = v
						index_std = index_std + 1
					end
   			end
			end
		end
	end

	-- Search for the closest aggressor -----------
	if result == 0 then
		-- Log("No mob to avoid")
   	result = GetClosest(enemy_1st)
   end
	if result == 0 then
		-- Log("No primary target in range")
		result = GetClosest(enemy_std)
	end
	if result == 0 then
		-- Log("No standard target in range")
		result = GetClosest(enemy_lss)
	end
	if result == 0 then
		-- Log("No low priority target in range")
		local nearest_weak = GetClosest(enemy_wkk)
		if (nearest_weak ~= 0) and (MyState == IDLE_ST) then
			if (Self_Per_HP <= Config[Self].HP_PERC_SAFE2ATK) then
				-- No real threats were found, but some weak (= BEHA_attack_weak) enemies are hitting me: usually the
				-- aggressive scan (GetMyEnemy_AnyNoKS function) takes care of them (in the correct priority order), but
				-- now the homunculus HPs are under the safe level, so he/she is in defensive-only mode: no agressive
				-- scan will be started at this time! So it's ok to attack now this low priority monster
				result = nearest_weak
			end
		end
	end
	if result == 0 then
		-- Log("No (serious) enemy is attacking me")
	else
		Log(string.format("<defensive scan> Target found (ID %d):", result))
		if result < 50000 then 
			HTact = GetFullTact_LoID(GetMobIDFromActor(result))
		else
			HTact = GetFullTact(GetMobIDFromActor(result))
		end
	end

	-- Log("---- END OF DEFENSIVE SCAN")
	return result
end

--------------------------------------------------
function GetMyEnemy_AnyNoKS(myid)
--------------------------------------------------
	-- Log(string.format("---- AGGRESSIVE SCAN STARTED for ID (%d)", myid))
	local result = 0
	local actors = GetActors ()
	local players = {}
	local player_idx = 1
	local enemy_1st = {} -- primary targets
	local index_1st = 1
	local enemy_std = {} -- normal targets
	local index_std = 1
	local enemy_lss = {} -- last targets
	local index_lss = 1
	local mob_type, mob_target, mobBehav, ListedType

	for i,v in ipairs(actors) do -- fill the (not "friend") player list
		if (v ~= OwnerID and v ~= MyID) then
			if (0 == IsMonster(v)) then
				if isNotFriend(v) then
					players[player_idx] = v
					player_idx = player_idx + 1
				end
			end
		end
	end

	for i,v in ipairs(actors) do
		if (v ~= OwnerID) and (v ~= MyID) then 	  -- owner and homunculus aren't valid target, of course
			if (IsMonster(v) == 1) and (BlockedPathToEnemyID ~= v) then -- if target is a (reachable) monster

				local isOKToAttack = false
				mob_type = GetMobIDFromActor(v)		  -- get tact data
				if v < 50000 then
						mobBehav = GetTact_LoID(V_TACTBEHAV, mob_type)
					else
						mobBehav = GetTact(V_TACTBEHAV, mob_type)
					end
				if( (mobBehav == BEHA_attack_1st)
				or  (mobBehav == BEHA_attack)
				or  (mobBehav == BEHA_attack_last)
				or  (mobBehav == BEHA_attack_weak)
				) then										  -- OK, this a type of monster to attack
					if isMobMotionOK(v, players) then
-----------------------------------------------------------------------------------------------------------------------
						if MVP_KS_OK[mob_type] ~= nil then
--							Log("MVP. Verificao de KS ignorada.")
							isOKToAttack = true
						else
-----------------------------------------------------------------------------------------------------------------------
							mob_target = GetV(V_TARGET, v)  -- check for the monster's target
							if (mob_target == 0) then	-- if the monster is not in battle
								isOKToAttack = true	-- he/she can be attacked, unless...
								for i2,v2 in ipairs(players) do
									if (v == GetV(V_TARGET, v2)) then
										isOKToAttack = false   -- ...another player is already aiming at him/her (anti-KS)
										break
									end
								end
							else									  -- else, if the monster is already in battle
							------ funcao modificada
								local bool_temp = false
								if (0 ~= IsMonster(mob_target)) then
									bool_temp = true
								elseif (mob_target < 100000) then
									if (GetV(V_POSITION, mob_target) == -1) then
										bool_temp = true
									end
								end								
								
								if (mob_target == MyID)
								or (mob_target == OwnerID)
								or (not isNotFriend(mob_target))
								or bool_temp then
									isOKToAttack = true		  -- but with me/my owner/a friend, it's OK (anti-KS)
							-------------------------------------------------------------------------------
								end
							end
-----------------------------------------------------------------------------------------------------------------------
						end
-----------------------------------------------------------------------------------------------------------------------
					end
				end

				if isOKToAttack then
					if (mobBehav == BEHA_attack_1st) then
						enemy_1st[index_1st] = v
						index_1st = index_1st + 1
						-- Log(string.format(" added to (1st) as enemyID %d", v))
					elseif (mobBehav == BEHA_attack) then
						enemy_std[index_std] = v
						index_std = index_std + 1
						-- Log(string.format(" added to (std) as enemyID %d", v))
					else -- BEHA_attack_last, BEHA_attack_weak
						enemy_lss[index_lss] = v
						index_lss = index_lss + 1
						-- Log(string.format(" added to (lss) as enemyID %d", v))
					end
				end

			end
		end
	end

	-- Searches for nearest target ---------------
	result = GetClosest(enemy_1st)
	if (result == 0) then
		-- Log("No primary target in range")
		result = GetClosest(enemy_std)
	end
	if (result == 0) then
		-- Log("No standard target in range")
		result = GetClosest(enemy_lss)
	end
	if (result == 0) then
		-- Log("No target in range")
	else
		Log(string.format("<aggressive scan> Target found (ID %d):", result))
		if result < 50000 then 
			HTact = GetFullTact_LoID(GetMobIDFromActor(result))
		else
			HTact = GetFullTact(GetMobIDFromActor(result))
		end
	end

	BlockedPathToEnemyID = 0
	-- Log("---- END OF AGGRESSIVE SCAN")
	return result
end

--------------------------------------------------
function GetClosest(EnemiesOnSight)
--------------------------------------------------
	local result = 0
	local nearest_dst = 100
	local owner_dst = 0
	local dst = 0
	for i,v in ipairs(EnemiesOnSight) do
		dst = GetDistance2(MyID, v)
		-- Log(string.format("distance from EnemyID %d is %d", v, dst))
		owner_dst = GetDistance2(OwnerID, v)
		if (dst < nearest_dst and owner_dst < Config[Self].TOO_FAR_TARGET) then
			result = v
			nearest_dst = dst
			-- Log(string.format("he/she is now the closest (%d)", nearest_dst))
		end
	end
	return result
end

--------------------------------------------------
function isCloseToOwner()
--------------------------------------------------
	-- OWNER_CLOSEDISTANCE is no longer used
	return (math.abs(OwnerX - MyX) <= 2) and (math.abs(OwnerY - MyY) <= 2)
end

--------------------------------------------------
function limMove(ID, x, y) -- Movement protection filter
--------------------------------------------------
	if x == -1 or y == -1 then return; end -- not valid coordinates
	if GetDistance(OwnerX, OwnerY, x, y) < Config[Self].TOO_FAR_TARGET then
		if IsFevDanc then
			Log("...Dancing... (state "..MyState..") ("..x..","..y..")")
		else
			Log("...Moving... (state "..MyState..") ("..x..","..y..")")
		end
		stdMove(ID, x, y)
	else
		Log("[ -> IDLE_ST] Destination was too far: back to owner")
		MyState = IDLE_ST
		MyEnemy = 0
		MySkill = 0
		IdleStartTime = GetTick()
        MoveToOwner(MyID)
	end
end

function ReturnToOwner(ID)
	local x, y = GetV(V_POSITION, OwnerID)
	if x == -1 or y == -1 then 
		stdMoveToOwner(ID)
		return
	end 
	if I_AM_MERC then
		stdMove(ID, x - 1, y)
	else
		stdMove(ID, x + 1, y)
	end
end

--------------------------------------------------
function AI_Sec(myid)
--------------------------------------------------
	MyID = myid
	local msg = GetMsg (myid) -- command
	local rmsg = GetResMsg (myid) -- reserved command

	-----------------------------------------------
	if not Initialized then -- AI init. section
	-----------------------------------------------
		Log("===INIT START===")
		if MyID > 50000 then
			Private_Srv = true
		end
		OwnerID = GetV(V_OWNER, MyID)
		LoadSkillsForMercenary()
		local SaveMyID = io.open(Config[Self].FILE_MYID, "w")
		if SaveMyID ~= nil then
			SaveMyID:write(MyID.."\n")
			SaveMyID:close()
		end
		FriendList_Load()
		stdMove = Move; Move = limMove -- replace Move with the protected version
		stdMoveToOwner = MoveToOwner; MoveToOwner = ReturnToOwner
		StartupTime = GetTick()
		IdleStartTime = StartupTime
		if FileExists(F_FOLLOW_LCK) then
			MyState = FOLLOW_CMD_ST
		end
		ModInit() -- initialize the mod too
		Log("===INIT DONE===")
		Initialized = true
	end
	-----------------------------------------------
	
	-----------------------------------------------------
	-- Atualizando variaveis globais de status
	-----------------------------------------------------
	Self_HP     = GetV(V_HP, MyID)
	Self_Max_HP = GetV(V_MAXHP, MyID)
	Self_Per_HP = (Self_HP/Self_Max_HP)*100
	Self_SP     = GetV(V_SP, MyID)
	Self_Max_SP = GetV(V_MAXSP, MyID)
	Self_Per_SP = (Self_SP/Self_Max_SP)*100
	Self_Motion = GetV(V_MOTION, MyID)
	Mast_HP     = GetV(V_HP, OwnerID)
	Mast_Max_HP = GetV(V_MAXHP, OwnerID)
	Mast_Per_HP = (Mast_HP/Mast_Max_HP)*100
	Mast_Motion = GetV(V_MOTION, OwnerID)
	Scre_Actors = GetActors()
	-----------------------------------------------------
	
	MyX, MyY = GetV(V_POSITION, MyID)
	OwnerX, OwnerY = GetV(V_POSITION, OwnerID)
	
	if (MyX == OwnerX) and (MyY == OwnerY) and ((MyState == IDLE_ST) or (MyState == FOLLOW_CMD_ST)) then
		Move(MyID, MyX+1, MyY+1)
		MyX = MyX+1
		MyY = MyY+1
	end
	
	if msg[1] == NONE_CMD then
		if rmsg[1] ~= NONE_CMD then
			if List.size(ResCmdList) < 10 then
				List.pushright (ResCmdList,rmsg)
			end
		end
	else
		List.clear (ResCmdList)
		ProcessCommand (msg)
	end
	
	if not ((MyState == ATTACK_ST) and (AI_ent_p - AtkStartTime > 20000)) then
		UpdateMobID(false)
		AddAllyAIToFriends(false)
	end
	
	if MyEnemy ~= CanDoKS then CanDoKS = 0 end
	
	if (MyEnemy == MyID) or (MyEnemy == OwnerID) or (Friends[MyEnemy] ~= nil) then
		StopHere("[-> IDLE_ST] Alvo invalido selecionado para atacar.")
	end
	
 	if 	(MyState == IDLE_ST) then
		DoAvailableSpecialSkills()
		OnIDLE_ST()
	elseif (MyState == CHASE_ST) then
		OnCHASE_ST()
	elseif (MyState == EVADE_ST) then
		OnEVADE_ST()
	elseif (MyState == ATTACK_ST) then
		if Config[Self].CONFG_VAR_DANC == 1 then
			if (Self_Per_SP > Config[Self].CONFG_VAR_DANC_SP) and 
			   (Self_Per_HP > Config[Self].CONFG_VAR_DANC_HP) then
				Log("AI(): Dancing... ")
				local zx, zy
				if MercType < 11 then
					zx, zy = GetV(V_POSITION, MyID)
				else
					zx, zy = GetV(V_POSITION, MyEnemy)
				end
				OnATTACK_ST ()
				Move(MyID,zx+1,zy)
				Move(MyID,zx-1,zy)
				OnATTACK_ST ()
				Move(MyID,zx,zy+1)
				Move(MyID,zx,zy-1)
			else
				OnATTACK_ST()
			end			
		else
			OnATTACK_ST()
		end
	elseif (MyState == FOLLOW_ST) then
		DoAvailableSpecialSkills()
		OnFOLLOW_ST()
	elseif (MyState == BUGPOSI_ST) then
		OnBUGPOSI_ST()
	-----------------------------------
	elseif (MyState == OUTMOVE_NEWFRIEND_ST) then
		OnOUTMOVE_NEWFRIEND_ST()
	elseif (MyState == OUTMOVE_DELFRIEND_ST) then
		OnOUTMOVE_DELFRIEND_ST()
	-----------------------------------
	elseif (MyState == MOVE_CMD_ST) then
		OnMOVE_CMD_ST()
	elseif (MyState == STOP_CMD_ST) then
		OnSTOP_CMD_ST()
	elseif (MyState == ATTACK_OBJECT_CMD_ST) then
		OnATTACK_OBJECT_CMD_ST ()
	elseif (MyState == ATTACK_AREA_CMD_ST) then
		OnATTACK_AREA_CMD_ST()
	elseif (MyState == PATROL_CMD_ST) then
		DoAvailableSpecialSkills()
		OnPATROL_CMD_ST()
	elseif (MyState == HOLD_CMD_ST) then
		OnHOLD_CMD_ST()
	elseif (MyState == SKILL_OBJECT_CMD_ST) then
		OnSKILL_OBJECT_CMD_ST()
	elseif (MyState == SKILL_AREA_CMD_ST) then
		OnSKILL_AREA_CMD_ST()
	elseif (MyState == FOLLOW_CMD_ST) then
		DoAvailableSpecialSkills()
		OnFOLLOW_CMD_ST()
	end
end
